//
//	Copyright  Nemetschek Vectorworks, Inc.
//	Use of this file is governed by the Nemetschek Vectorworks SDK License Agreement
//	http://developer.vectorworks.net/index.php?title=Vectorworks_SDK_License
//

#include <Python.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <vector>

// Windows includes
#ifdef WIN32    
	#include <process.h>
	HANDLE handle;
// POSIX includes
#else    
	#include <pthread.h>
	pthread_t mythread;
#endif

using namespace VectorWorks::Filing;
	
// ==========================================================================
void ThreadProc(void*);

// ==========================================================================
TXString GetVTGEscapedAppFolderPath()
{
	IApplicationFoldersPtr appFolderPtr(IID_ApplicationFolders);
	IFolderIdentifierPtr folderID;
	appFolderPtr->GetApplicationFolder(&folderID);
	TXString appFolder;
	folderID->GetFullPath(appFolder);

	char* slash = "\\";
	char* doubleSlash = "\\\\";
	appFolder.Replace(slash, doubleSlash);

	return appFolder;
}

// ==========================================================================
TXString GetVTGEscapedPythonSourcePath()
{
	return TXString(GetVTGEscapedAppFolderPath() + "PythonSource\\\\VTG_Prototype.py");
}

// ==========================================================================
TXString GetVTGEscapedProjectPath()
{
	return TXString(GetVTGEscapedAppFolderPath() + "VTG_Projects\\\\");
}

// ==========================================================================
TXString GetVTGEscapedLogPath()
{
	return TXString(GetVTGEscapedAppFolderPath() + "Logs\\\\");
}

// ==========================================================================
// To Do: Tweak return values
int CallPythonFunction(PyObject* pFunc, std::vector<const char*> params)
{
	PyObject *pValue, *pArgs;

	int numParams = params.size();

	pArgs = PyTuple_New(numParams);

	// Prepare arguments
	for (int i = 0; i < numParams; ++i) {
		pValue = PyString_FromString(params[i]);
					
		if (!pValue) {
			PyErr_Print();
			return 1;
		}
					
		PyTuple_SetItem(pArgs, i, pValue);
	}
            
	// Call function
	pValue = PyObject_CallObject(pFunc, pArgs);

	// Handle errors
	if (pValue != NULL) {
		int returnValue = PyInt_AsLong(pValue);
		
		if (returnValue == -1) {
			gSDK->AlertCritical("PyInt_AsLong cast failed.");
		}

		else if (returnValue != 0) {
			PStr255	strReturnValue = "";
#if 1
			TXString message = "Error executing Python script.";
#else
			::NumToString(returnValue, strReturnValue);

			TXString message = "Error executing Python script (Error ";
			message << strReturnValue;
			message << ").";
#endif
			message << "  See log file for more details: ";

			TXString advice = "";

			if (VERIFYN(kDonne, params.size() > 0)) {
				advice = params[0];
			}

			gSDK->AlertCritical(message, advice);
		}
	}

	// Cleanup
	if (pArgs != NULL) {
		Py_DECREF(pArgs);
	}

	if (pValue != NULL) {
		Py_DECREF(pValue);
	}
	else {
		PyErr_Print();
	}

	return 0;
}

// ==========================================================================
// To Do: Tweak return values
int PythonWrapperMultiThreaded(const char* functionName, std::vector<const char*> params)
{
    PyObject *pName, *pModule, *pDict, *pFunc;
    PyThreadState *mainThreadState, *myThreadState;
    PyInterpreterState *mainInterpreterState;
	
    // Initialize the Python Interpreter
    Py_Initialize();

    // Initialize thread support
    PyEval_InitThreads();

    // Save a pointer to the main PyThreadState object
    mainThreadState = PyThreadState_Get();

	// release the lock
	PyEval_ReleaseLock();

	// get the global lock
	PyEval_AcquireLock();

	// get a reference to the PyInterpreterState
	mainInterpreterState = mainThreadState->interp;

	// create a thread state object for this thread
	myThreadState = PyThreadState_New(mainInterpreterState);

	// free the lock
	PyEval_ReleaseLock();

	// grab the global interpreter lock
	PyEval_AcquireLock();

	// swap in my thread state
	PyThreadState_Swap(myThreadState);

    // Build the name object
    pName = PyString_FromString("VTG_Prototype");

    // Load the module object
    pModule = PyImport_Import(pName);

    // pDict is a borrowed reference 
    pDict = PyModule_GetDict(pModule);

    // pFunc is also a borrowed reference 
    pFunc = PyDict_GetItemString(pDict, functionName);

	if (PyCallable_Check(pFunc)) {
		CallPythonFunction(pFunc, params);
	}

    // Clean up
    Py_DECREF(pModule);
    Py_DECREF(pName);

	// clear the thread state
	PyThreadState_Swap(NULL);

	// release our hold on the global interpreter
	PyEval_ReleaseLock();

	// grab the lock
	PyEval_AcquireLock();

	// swap my thread state out of the interpreter
	PyThreadState_Swap(NULL);

	// clear out any cruft from thread state object
	PyThreadState_Clear(myThreadState);

	// delete my thread state object
	PyThreadState_Delete(myThreadState);

	// release the lock
	PyEval_ReleaseLock();

	// shut down the interpreter
	PyEval_AcquireLock();
	PyThreadState_Swap(mainThreadState);
	Py_Finalize();

    return 0;
}

// ==========================================================================
// To Do: Tweak return values
int PythonWrapper(const char* moduleName, const char* functionName, std::vector<const char*> params)
{
    PyObject *pName, *pModule, *pDict, *pFunc;

    // Initialize the Python Interpreter
    Py_Initialize();

    // Build the name object
    pName = PyString_FromString(moduleName);
	if ( pName )
	{
		// Load the module object
		pModule = PyImport_Import(pName);
		if ( pModule )
		{
			// pDict is a borrowed reference 
			pDict = PyModule_GetDict(pModule);
			if ( pDict )
			{
				// pFunc is also a borrowed reference 
				pFunc = PyDict_GetItemString(pDict, functionName);

				if (pFunc && PyCallable_Check(pFunc)) {
					CallPythonFunction(pFunc, params);
				}
			}

			// Clean up
			Py_DECREF(pModule);
		}
		else
		{
			PyObject *exc, *val, *tbk;
			PyObject *exc_str;

			DMSG((kVStanev,"error in running script\n"));
			if( PyErr_Occurred())
			{
				if(PyErr_ExceptionMatches(PyExc_Exception))
				{
					printf("exception received in C\n");
				}
				PyErr_Fetch(&exc, &val, &tbk);
				exc_str = PyObject_Str(exc);
				DMSG((kVStanev,"exception received: %s\n", PyString_AsString(exc_str)));
				DMSG((kVStanev,"exception value: %s\n",PyString_AsString(val)));

				TXString msg;
				msg.Format( "Python exception: %s : %s", PyString_AsString(exc_str), PyString_AsString(val) );
				MessageBox( NULL, msg, "PYTHON", 0 );

				Py_DECREF(exc_str);
				Py_DECREF(exc);
				Py_DECREF(val);
				Py_DECREF(tbk);
			}
		}

		Py_DECREF(pName);
	}

    // Finish the Python Interpreter
    Py_Finalize();

    return 0;

}

#if 0

// ==========================================================================
void PythonWrapperMultiThreadedSetup()
{
    PyThreadState *mainThreadState, *myThreadState;
    PyInterpreterState *mainInterpreterState;
	
    // Initialize the Python Interpreter
    Py_Initialize();

    // Initialize thread support
    PyEval_InitThreads();

    // Save a pointer to the main PyThreadState object
    mainThreadState = PyThreadState_Get();

	// release the lock
	PyEval_ReleaseLock();

	// get the global lock
	PyEval_AcquireLock();

	// get a reference to the PyInterpreterState
	mainInterpreterState = mainThreadState->interp;

	// create a thread state object for this thread
	myThreadState = PyThreadState_New(mainInterpreterState);

	// free the lock
	PyEval_ReleaseLock();

	// grab the global interpreter lock
	PyEval_AcquireLock();

	// swap in my thread state
	PyThreadState_Swap(myThreadState);
}

// ==========================================================================
void PythonWrapperMultiThreadedSetdown()
{
	// clear the thread state
	PyThreadState_Swap(NULL);

	// release our hold on the global interpreter
	PyEval_ReleaseLock();

	// grab the lock
	PyEval_AcquireLock();

	// swap my thread state out of the interpreter
	PyThreadState_Swap(NULL);

	// clear out any cruft from thread state object
	PyThreadState_Clear(myThreadState);

	// delete my thread state object
	PyThreadState_Delete(myThreadState);

	// release the lock
	PyEval_ReleaseLock();

	// shut down the interpreter
	PyEval_AcquireLock();
	PyThreadState_Swap(mainThreadState);
	Py_Finalize();
}

#endif