//
//	Copyright  Nemetschek Vectorworks, Inc.
//	Use of this file is governed by the Nemetschek Vectorworks SDK License Agreement
//	http://developer.vectorworks.net/index.php?title=Vectorworks_SDK_License
//

#pragma once

#include "DirectModellingSupport.h"
#include "WidgetsEditProvider.h"

namespace VWFC
{
	namespace PluginSupport
	{
		using namespace VectorWorks::Extension;

		// ------------------------------------------------------------------------------------------------------
		class VWParametric_EventSink	: public VCOMImpl<ILegacyEventSink>
		{
		public:
									VWParametric_EventSink(IVWUnknown* parent);
			virtual					~VWParametric_EventSink();

			void					SetUniversalName(const TXString& name);

		// Plug-in Object action/events constants 
		public:
			// initialization called on the first event
			virtual void			InitializeExt();
			// deitilialization, called on destructor time
			// (you can just use the destructor)
			virtual void			DeinitializeExt();
			// The Plug-in Object code should re-create all its component objects using current parameter values.
			virtual EObjectEvent	Recalculate();
			// Object Definition Procedure (ODP) may optionally present its own preferences dialog, or let VW pose the default dialog.
			virtual EObjectEvent	Preference();

		public:
			// The Plug-in Object may add extended properties.
			virtual EObjectEvent	OnInitXProperties(CodeRefID objectID);
			// The Plug-in Object called with kParametricOnMove when kHasMoveDependancy extended property set
			virtual EObjectEvent	OnSpecialEditID();
			// The user has pressed a button in the Properties UI of an object
			virtual EObjectEvent	OnObjectUIButtonHit(long buttonID);
			// Update object's shape pane widgets
			virtual bool			OnWidgetUpdate(IWidgetsProvider* pWidgetProvider);
			// Handle widget change
			virtual EObjectEvent	OnAddState(ObjectState& stateInfo);
            // The user has manipulated one of plugin object's control points.
            virtual long            OnCursorEvent(ObjectEventID action, long message);
			// Obtain the tool name that is to be activated for this object
			// when doing 'create similar object' command
			virtual TXString		OnGetToolName() const;
			// Object context menu support
			virtual void			OnContextMenuInit(IContextMenuProvider* pContextMenuProvider);
			virtual void			OnContextMenuEvent(const ObjectContextMenuEvent& eventData);
			// Eyedropper notifies that it is going to put the record 'hObjRecordCopy' as record for the current parametric object
			// In this event the object can alter the copy to ensure correct data transfer (the data has already been copied to the record)
			// NOTE! This object's record is still in place when this event is called
			virtual void			OnEyedropperPrepareCopy(MCObjectHandle hObjRecordCopy);
			// Eyedropper notification that has finished with copying this object
			virtual void			OnEyedropperAfterCopy(MCObjectHandle srcObject);
			// suply specific geometry
			virtual MCObjectHandle	OnGetSpecificGeometry(ObjectGetSpecificGeometryCall::ESpecifier geomType);

			// The default event notificaiotn
			// if no virutal is present for an event, this function is called to allow the user to catch everything
			virtual long			OnDefaultEvent(long action, long message);

		// ILegacyEventSink
		public:
			virtual long VCOM_CALLTYPE			LegacyMain(long action, long message1, long message2);

		private:
			// DO NOT USE THESE!!!
			virtual void			GetUserData(void) const;
			virtual void			SetUserData(void) const;

		protected:
			MCObjectHandle		fhObject;
			TXString			fUnivName;
			bool				fbInitialized;
		};

		// ------------------------------------------------------------------------------------------------------
		struct SParametricDef
		{
			SSTRResource		fLocalizedName;
			EParametricSubType	fSubType;
			bool				fResetOnMove;
			bool				fResetOnRotate;
			bool				fWallInsertOnEdge;
			bool				fWallInsertNoBreak;
			bool				fWallInsertHalfBreak;
			bool				fWallInsertHideCaps;
		};

		// ------------------------------------------------------------------------------------------------------
		struct SParametricParamChc
		{
			short				fChcID;
			const char*			fUniversalChoice;
			SSTRResource		fLocalizedChoice;
		};

		// ------------------------------------------------------------------------------------------------------
		struct SParametricParamDef
		{
			const char*				fUniversalName;
			SSTRResource			fLocalizedName;
			const char*				fDefaultImperial;
			const char*				fDefaultMetric;
			EFieldStyle				fFieldStyle;
			short					fChoicesChcID;		// mathing SParametricParamChc::fChcID
		};

		// ------------------------------------------------------------------------------------------------------
		class VWExtensionParametric	: public VCOMImmediateImpl<IExtensionParametric>
		{
		public:
											VWExtensionParametric(	CallBackPtr cbp,
																	const SParametricDef& parametricDef,
																	const SParametricParamDef* arrParamDefs,
																	const SParametricParamChc* arrParamChcs = NULL);
			virtual							~VWExtensionParametric();


		// IExtension
		public:
			virtual IEventSink* VCOM_CALLTYPE		QueryEventSink(const TSinkIID& iid);

		// IExtensionParametric
		public:
			virtual EParametricSubType VCOM_CALLTYPE	GetObjectType();
			virtual const TXString& VCOM_CALLTYPE		GetLocalizedName();
			virtual void VCOM_CALLTYPE					GetResetModeOnMove(bool& outOnMove, bool& outOnRotate);
			virtual void VCOM_CALLTYPE					GetWallInsertMode(bool& outOnEdge, bool& outNoBreak, bool& outHalfBreak, bool& outHideCaps);

			

		protected:
			virtual void VCOM_CALLTYPE			SetLocalizedParamDef( const short listID );
			virtual void VCOM_CALLTYPE			SetLocalizedParamChc( const short listID );

			virtual void						OnRefCountZero();
			virtual ILegacyEventSink*			CreateLegacyEventSink(IVWUnknown* parent) = 0;
			virtual bool						GetProtectionSinkIID(TSinkIID& outIID);
			virtual IProviderProtection*		CreateProtectionSink(IVWUnknown* parent);
			virtual void						DefineSinks();
			template<class T> void				DefineSink(const TSinkIID& sinkIID);

		private:
			CallBackPtr						fCBP;

			typedef std::map<SSTRResource, TXString>		TResStringCache;
			typedef std::map<short, const SParametricParamChc*>	TChcCache;

			const SParametricDef&			fParametricDef;
			TXString						fstrLocalizedName;
			TResStringCache					fResStringChache;
			size_t							fParamsCount;
			const SParametricParamDef*		farrParamDefs;
			const SParametricParamChc*		farrParamChcs;
			TChcCache						fParamChcCache;
			TXString						fEmptyString;

			class CParametersProvider;

			CParametersProvider*			fParametersSink;
			ILegacyEventSink*				fLegacyEventSink;
			bool							fProtectionObtained;
			bool							fHasProtection;
			TSinkIID						fProtectionSinkIID;
			IProviderProtection*			fProtectionSink;

			typedef std::map<TSinkIID, IEventSink*> TSinksMap;
			TSinksMap						fmapSinks;
			TSinkIID						fDefinedSinkIID;
			IEventSink*						fDefinedSink;

		private:
			// -------------------------------------------------------------------
			class CParametersProvider : public VCOMImpl<IParametricParamsProvider>
			{
			public:
						CParametersProvider(VWExtensionParametric* parent);
				virtual	~CParametersProvider();

			// IParametricParamsProvider
			public:
				virtual size_t VCOM_CALLTYPE			GetParamsCount();
				virtual void VCOM_CALLTYPE				GetParamAt(size_t paramIndex, TXString& outUniversalName, EFieldStyle& outType);
				virtual const TXString& VCOM_CALLTYPE	GetParamNameAt(size_t paramIndex);
				virtual void VCOM_CALLTYPE				GetParamDefValueImperial(size_t paramIndex, TXString& outValue);
				virtual void VCOM_CALLTYPE				GetParamDefValueMetric(size_t paramIndex, TXString& outValue);
				virtual void VCOM_CALLTYPE				GetParamChoices(size_t paramIndex, TXStringArray& outArrUniversalNames, TXStringArray& outArrDisplayNames);

			private:
				void	GetDefaultValue(const char* szDefaultValue, TXString& outValue);

			private:
				VWExtensionParametric*	fParent;
			};
		};
	}

	namespace PluginSupport
	{
		template<class T> void VWExtensionParametric::DefineSink(const TSinkIID& sinkIID)
		{
			if ( fDefinedSink == NULL && sinkIID == fDefinedSinkIID ) {
				T*	instance	= new T( this );
				fDefinedSink	= instance;
			}
		}
	}
}

