//
//	Copyright  Nemetschek Vectorworks, Inc.
//	Use of this file is governed by the Nemetschek Vectorworks SDK License Agreement
//	http://developer.vectorworks.net/index.php?title=Vectorworks_SDK_License
//

#pragma once

#include <stddef.h>

typedef unsigned char		Uint8;
typedef	unsigned short		Uint16;
typedef	unsigned long		Uint32;
#if _WINDOWS
typedef unsigned __int64	Uint64;
#else	// Mac
typedef unsigned long long	Uint64;
#endif

#ifndef Byte
typedef unsigned char Byte;
#endif

#ifndef Ptr
typedef char *                          Ptr;
#endif
#ifndef Handle
typedef Ptr *                           Handle;
#endif
#ifndef Size
typedef long                            Size;
#endif

#ifndef SInt16
typedef signed short                    SInt16;
#endif

#ifndef OSErr
typedef SInt16                          OSErr;
#endif

#if _WINDOWS
#ifdef VWMM_EXPORTS
	#define MM_DLL_EXPORT __declspec (dllexport)
#else	// not VWMM_EXPORTS
	#define MM_DLL_EXPORT __declspec (dllimport)
#endif	// VWMM_EXPORTS
#else	// not _WINDOWS
	#define MM_DLL_EXPORT
#endif

#ifndef _WINDOWS
// Declare Mac calls for thread safety
struct OpaqueMPCriticalRegionID;
typedef struct OpaqueMPCriticalRegionID*  MPCriticalRegionID;
#endif

// -------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------
#include "VWMM_Defines.h"
#include "VWMM_Memblock.h"
#include "VWMM_Pool.h"

// -------------------------------------------------------------------------------------------------
// Used for saving the stack in allocation calls
// -------------------------------------------------------------------------------------------------
#pragma pack(push, 1)
struct StackTraceEntry
{
	StackTraceEntry *next;
	int numCalls;
	Uint32 averageSize;
	char stack[MEMORY_BLOCK_STACKTRACE_SIZE];
};
#pragma pack(pop)

// -------------------------------------------------------------------------------------------------
// VWMM encapsulates our local memory management system.
// -------------------------------------------------------------------------------------------------
namespace VWMM
{

#define PARTITION_IDENTIFIER_BUFFER_SIZE (128)
// -------------------------------------------------------------------------------------------------
// A partition is a modular, self contained area  that contains all the information for an instance of VWMM.
// These were stored in globals before 
// -------------------------------------------------------------------------------------------------
class MM_DLL_EXPORT Partition
{
public:
	enum
	{
		PARTITION_FLAG_USEPOOL = 1,	// if set, new allocations may use the pool. If clear, they won't.
	};
protected:
	Handle fPool;	// the pool for this partition. May be NULL if no pool has been set up.
	Uint32 fFlags;	// flags for this partition, see above.

	MemoryBlock *fAllBlocks[MEMORYBLOCK_HASH_SIZE];	// all blocks currently in use in this partition (including pooled)
	int fBlockListCounts[MEMORYBLOCK_HASH_SIZE];	// counts the amount of blocks in each list
	
	// All resources (Created with GetOrCreateHandleFromQTResourceHandle()) are stored in these lists.
	MemoryBlock *fAllResourceBlocks[MEMORYBLOCK_HASH_SIZE];
	int fResourceBlockListCounts[MEMORYBLOCK_HASH_SIZE];
	
	int fBytesAllocated;		// total bytes currently allocated (as the user sees, does not include header info)
	int fNumBlocks;				// total number of blocks allocated (should be in sync with contents of fAllBlocks)
	unsigned int fNumCalls;		// this just counts the number of calls to allocate/resize for printing out debug info intermittantly
	unsigned int fNumPoolHits;	// increased when memory is allocated from pool instead of from heap
	unsigned int fNumAllocatesTotal;	// total amount of allocates.

	int fNumCallsBetweenOutput;	// how often to print out debug info
	unsigned int fMagicNumber;	// used to sanity check
	char fIdentifier[PARTITION_IDENTIFIER_BUFFER_SIZE];	// used to retrieve a partition from the map

#if MEMORY_BLOCK_STACKTRACE_HISTORY	// format for each is void * next, stacktrace string
	StackTraceEntry *fStackTraceHistory;
#endif // MEMORY_BLOCK_STACKTRACE_HISTORY

	// each slot x will contain the sum of allocations from (gAllocSizesLessThan[x - 1] + 1) to gAllocSizesLessThan[x] (for example slot 1 will be sizes 11 - 50).
	unsigned int fAllocSizesLessThanOrEqual[NUM_ALLOCSIZE_SLOTS];	// This holds the max alloc size for the given slot.
	unsigned int fAllocSizeCounts[NUM_ALLOCSIZE_SLOTS];				// this holds the count for the given slot
	int fAllocSizeTotals;	// count for average of above

	unsigned int MemoryStat_GetSlot(int size);	// given a memory size, return which slot the stat should go into.
	void MemoryStat_GetSlotBoundsStr(char *str, int slotNum); // given a slot and a buffer, fill buffer with a string such as "100-250" based on the range stored in slot
	void MemoryStat_AddAlloc(int size);	// add the given size allocation to the statistics

	// Some utility/helper functions
	void HookupBlockToList(MemoryBlock *block, MemoryBlock **blockList, int *countArray);		// assumes block valid and not in a list.
	void RemoveBlockFromList(MemoryBlock *block, MemoryBlock **blockList, int *countArray);	// assumes block is in this list.

	// Used to keep a mapping of partitions to be shared among all modules in a process
	static Partition *partitionList;
	void AddThisToList(void);			// assumes this doesn't have to be removed from a list.. only adds it.
	void RemoveThisFromList(void);
	Partition *fNextInList, *fPrevInList;

	bool BlockIsResource(MemoryBlock* block);
	MemoryBlock* AllocateFromExistingResource(void* mem, Uint32 size);	// creates a memory block that points to resource memory.

public:
	Partition();
	~Partition();

	// Retrieve the Partition with given ID. Potentially creating if necessary.
	static Partition *GetPartitionByIdentifier(char *identifier, bool createIfNotFound);
	static Partition *GetPartitionListHead(void);

	void SetIdentifier(char *newIdentifier);
	void GetIdentifier(char *buffer);	// Note that buffer must be at least PARTITION_IDENTIFIER_BUFFER_SIZE.

	Uint32 GetFlags(void) {return fFlags;}
	void SetFlags(Uint32 newFlags) {fFlags = newFlags;}

	Handle  GetCurrentPool(void) {return fPool;}
	void	SetCurrentPool(Handle poolHandle);		// set the current pool.

	// These functions are new in VWMM and provide generic memory operations. The QT replacements below will call these.
	void* Allocate(Uint32 size);	// basic allocation of memory block
	void Deallocate(void* mem);		// this must be called to free up the block created by Allocate.

	void* GetHandle(void* mem);								// given memory allocated by this library, return a handle.
	Uint32 GetVWMMHandleSize(void* handlein);				// given a handle, return its size

	bool PointerBelongsToUs(const void* mem);						// return true if the memory beloings to use. This is fairly expensive.
	bool HandleBelongsToUs(void* handle);
	bool PointerIsValid(void* mem, bool checkForOwnage);	// return true if the pointer is valid
	bool HandleIsValid(void* handle, bool checkForOwnage);	// return true if the handle is valid

	bool Resize(void* handle, Uint32  size);					// given a handle allocated by us, resize the memory
	void* DuplicateHandle(void* handlein);					// duplicate the passed handle.

	void* GetUserData(void* handlein);						// return the value of user data stored in the memory block header.
	void SetUserData(void* data, void* handlein);			// set the value of user data stored in the memory block header.

	// Functions dealing with Mac Resource handles
	// Note that if this can find an existing resource with the same qtHandle stored in it, it will just return that one similar to how Mac Resources do
	Handle GetOrCreateHandleFromQTResourceHandle(void** qtHandle, Uint32 size, short homeFileRef);
	void DeallocateResource(void* handlein);						// blocks allocated from AllocateFromExistingResource shoudl be deallocated with this call
	
	Handle FindBlockWithQTResourceHandle(void** qtHandle, short homeFileRef);	// attempt to find a block with the given handle
	Handle GetOriginalQTResource(void* handlein);			// return the QT resource handle if possible
	void SetQTResourcePtr(void* handlein, void** qtHandle, short homeFileRef);

	static void VWMM_Assert(bool condition, const char* reason);	// defined here so we can enable them in release mode if desired.

	static void *MemMem(const void *haystack, size_t hlen, const void *needle, size_t nlen);
	bool InsertData(void* handle, Uint32 offset, void* data, Uint32 dataSize);

	// These functions are replacements for calls into QTMM. They have the same signature as the original QT calls, but 
	// call into our local VWMM code instead.
	Handle		NewHandle(Size byteCount);
	Handle		NewHandleClear(Size byteCount);
	Ptr			NewPtr(Size byteCount);
	Ptr			NewPtrClear(Size byteCount);
	bool		SetHandleSize(Handle h, Size byteCount);
	bool		SetPtrSize(Ptr p, Size byteCount);
	bool		HandToHand(Handle &h);
	void		PtrToHand(const void *srcPtr, Handle &dstHndl, long size);
	bool		PtrToXHand(const void *srcPtr, Handle dstHndl, long size);
	bool		HandAndHand(Handle hand1, Handle hand2);
	bool		PtrAndHand(const void *p, Handle h, Size byteCount);
	Size		GetHandleSize(Handle h);
	Size		GetPtrSize(Ptr p);
	void		DisposePtr(Ptr p);
	void		DisposeHandle(Handle h);
	short		MemError(void);									// does nothing for gVWMM pointer. Here so that existing code can switch back to QTMM
	void		HLock(Handle h);	// this exists for backwards compatibility with handle lock/unlock mechanisms
	void		HUnlock(Handle h);	// this exists for backwards compatibility with handle lock/unlock mechanisms
	char		HGetState(Handle h);// this exists for backwards compatibility with handle lock/unlock mechanisms

	void		DeleteAllHandlesWithResourceFile(short resFile);

	// The following methods exist to provide a lock/unlock mechanism for older code that may require valid lock states
	bool		HandleIsLocked_Legacy(Handle h);
	bool		SetHandleLockState_Legacy(Handle h, bool newState);

	// The following functions are helpers to translate between the two memory systems.
	Handle	ConvertFromQTHandleToVWMMHandle(Handle qtHandle);	// given a QT handle return a VWMM handle copy
	Handle	ConvertFromVWMMHandleToQTHandle(Handle qtHandle);	// given a VWMM handle return a QT handle copy

	// Functions for Pooling
	Handle CreatePool(Uint32 maxBlockSize, Uint32 numBlocks);	// return a handle representing a VWMM pool with the given parameters, or NULL if we couldn't create one.
	Handle RequestPoolBlock(Handle pool);	// return a handle from the passed in Pool, or NULL. This will not fall back on standard calls if no pool blocks are available.
	void ReleasePoolBlock(Handle pool, Handle block);// release the block (which must be part of pool) back to the pool's pool of free blocks
	void DestroyPool(Handle pool);// delete this pool and all the memory inside it. You should make sure you're not still using any of that memory still.

	// Functions of a debug nature
	void PrintMemoryStats(Uint32 flags);
	static Uint32 GetMemoryBlockHash(MemoryBlock *block);
	
};	// class Partition

#ifndef _WINDOWS
	// Declare Mac calls for thread safety
	void EnterCriticalSection(MPCriticalRegionID *id);
	void LeaveCriticalSection(MPCriticalRegionID *id);
	
	void OutputDebugString(const char *str);
#endif // _WINDOWS
	
}	// namespace VWMM


// -------------------------------------------------------------------------------------------------
// These functions are direct calls into QuickTime - use only when necessary, for example, when they
// need to be passed back into other areas of Quicktime that expect native handles.
// -------------------------------------------------------------------------------------------------
namespace QTMM
{
	MM_DLL_EXPORT Handle	NewHandle(Size byteCount);
	MM_DLL_EXPORT Handle	NewHandleClear(Size byteCount);
	MM_DLL_EXPORT Ptr		NewPtr(Size byteCount);
	MM_DLL_EXPORT Ptr		NewPtrClear(Size byteCount);
	MM_DLL_EXPORT void		SetHandleSize(Handle h, Size byteCount);
	MM_DLL_EXPORT void		SetPtrSize(Ptr p, Size byteCount);
	MM_DLL_EXPORT OSErr		HandToHand(Handle *h);
	MM_DLL_EXPORT OSErr		PtrToHand(const void *srcPtr, Handle *dstHndl, long size);
	MM_DLL_EXPORT OSErr		PtrToXHand(const void *srcPtr, Handle dstHndl, long size);
	MM_DLL_EXPORT OSErr		HandAndHand(Handle hand1, Handle hand2);
	MM_DLL_EXPORT OSErr		PtrAndHand(const void *p, Handle h, Size byteCount);
	MM_DLL_EXPORT Size		GetHandleSize(Handle h);
	MM_DLL_EXPORT void		DisposePtr(Ptr p);
	MM_DLL_EXPORT void		DisposeHandle(Handle h);
	MM_DLL_EXPORT OSErr		MemError(void);
	MM_DLL_EXPORT void		HLock(Handle h);
	MM_DLL_EXPORT void		HUnlock(Handle h);
	MM_DLL_EXPORT char		HGetState(Handle h);
	MM_DLL_EXPORT bool		InsertData(Handle handle, Uint32 offset, void* data, Uint32 dataSize);
	
	

} // namespace QTMM

